// RB Whitaker excellent Using 3D Models tutorial (http://rbwhitaker.wikidot.com/using-3d-models) with BEPU physics integration
// Provided as an easy example to show a visual mesh and a physical entity linked together

// There is the most basic stuff here, no ground, no movable camera, etc so you'll just see a ship falling down

// Excercises for the reader: 
// -change gravity value and ship's mass and look at what's happening
// -add a ground inside the camera view area and look at collision
// -make the camera follow the ship all time
// -replace ship model with something else (your own model or some found on websites like www.turbosquid.com)
// -eventually add thruster from BEPU demo to the ship, or make the ship moves using some keys, etc

using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;

using BEPUphysics;
using BEPUphysics.Entities;
using BEPUphysics.CollisionShapes.ConvexShapes;



namespace Using3DModels
{
    /// <summary>
    /// This is the main type for your game
    /// </summary>
    public class Game1 : Microsoft.Xna.Framework.Game
    {
        GraphicsDeviceManager graphics;
        SpriteBatch spriteBatch;

        /// <summary>
        /// Stores the model that we are going to draw.
        /// </summary>
        private Model model;

        /// <summary>
        /// the physical object linked to the mesh (physics will be applied to this one)
        /// </summary>
        Entity convexHull;

        /// <summary>
        /// Stores the world matrix for the model, which transforms the 
        /// model to be in the correct position, scale, and rotation
        /// in the game world.
        /// </summary>
        private Matrix world = Matrix.CreateTranslation(new Vector3(0, 5, 0));

        /// <summary>
        /// Stores the view matrix for the model, which gets the model
        /// in the right place, relative to the camera.
        /// </summary>
        private Matrix view = Matrix.CreateLookAt(new Vector3(0, 0, 10), new Vector3(0, 0, 0), Vector3.UnitY);

        /// <summary>
        /// Stores the projection matrix, which gets the model projected
        /// onto the screen in the correct way.  Essentially, this defines the
        /// properties of the camera you are using.
        /// </summary>
        private Matrix projection = Matrix.CreatePerspectiveFieldOfView(MathHelper.ToRadians(45), 800f / 480f, 0.1f, 100f);

        /// <summary>
        /// The main simulator space
        /// </summary>
        Space space;


        public KeyboardState KeyboardInput;


        public Game1()
        {
            graphics = new GraphicsDeviceManager(this);
            Content.RootDirectory = "Content";
        }

        /// <summary>
        /// Allows the game to perform any initialization it needs to before starting to run.
        /// This is where it can query for any required services and load any non-graphic
        /// related content.  Calling base.Initialize will enumerate through any components
        /// and initialize them as well.
        /// </summary>
        protected override void Initialize()
        {
            graphics.PreferredBackBufferWidth = 1280;
            graphics.PreferredBackBufferHeight = 720;

            // initialize physics space and add a "standard" gravity
            space = new Space();
            space.ForceUpdater.Gravity = new BEPUutilities.Vector3(0, -9.81f, 0);

            base.Initialize();
        }

        /// <summary>
        /// LoadContent will be called once per game and is the place to load
        /// all of your content.
        /// </summary>
        protected override void LoadContent()
        {
            // Create a new SpriteBatch, which can be used to draw textures.
            spriteBatch = new SpriteBatch(GraphicsDevice);

            model = Content.Load<Model>("Ship");


            // convert the mesh into a BEPU usable entity (here the mesh is quite low-poly, so it's ok)
            // NOTE: this is not the best way to do it, using visual mesh as physical object can create performance issue
            BEPUutilities.Vector3[] staticTriangleVertices;
            int[] staticTriangleIndices;

            // ModelDataExtractor comes from BEPU demo
            ModelDataExtractor.GetVerticesAndIndicesFromModel(model, out staticTriangleVertices, out staticTriangleIndices);

            var convexHullShape = new ConvexHullShape(staticTriangleVertices);

            // mass set to 1 (kg) here, it could be anything, according to what you want to simulate
            // yes, it may be a toy plane :)
            // BE careful here to set both entities the SAME location ! (position here is for physical mesh, 
            // world Matrix above is the visual mesh location.)
            convexHull = new Entity(convexHullShape, 1) { Position = new BEPUutilities.Vector3(0, 5, 0) };
            space.Add(convexHull);
        }

        /// <summary>
        /// UnloadContent will be called once per game and is the place to unload
        /// all content.
        /// </summary>
        protected override void UnloadContent()
        {
        }

        /// <summary>
        /// Allows the game to run logic such as updating the world,
        /// checking for collisions, gathering input, and playing audio.
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        protected override void Update(GameTime gameTime)
        {
            KeyboardInput = Keyboard.GetState();

            // Allows the game to exit
            if ((GamePad.GetState(PlayerIndex.One).Buttons.Back == ButtonState.Pressed) || KeyboardInput.IsKeyDown(Keys.Escape))
            {
                this.Exit();
            }

            space.Update((float)gameTime.ElapsedGameTime.TotalSeconds);

            // update visual mesh location with physical entity location
            world = Matrix.CreateTranslation(new Vector3(convexHull.Position.X, convexHull.Position.Y, convexHull.Position.Z));

            base.Update(gameTime);
        }

        /// <summary>
        /// This is called when the game should draw itself.
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        protected override void Draw(GameTime gameTime)
        {
            GraphicsDevice.Clear(Color.CornflowerBlue);

            DrawModel(model, world, view, projection);

            base.Draw(gameTime);
        }

        /// <summary>
        /// Does the work of drawing a model, given specific world, view, and projection
        /// matrices.
        /// </summary>
        /// <param name="model">The model to draw</param>
        /// <param name="world">The transformation matrix to get the model in the right place in the world.</param>
        /// <param name="view">The transformation matrix to get the model in the right place, relative to the camera.</param>
        /// <param name="projection">The transformation matrix to project the model's points onto the screen correctly.</param>
        private void DrawModel(Model model, Matrix world, Matrix view, Matrix projection)
        {
            foreach (ModelMesh mesh in model.Meshes)
            {
                foreach (BasicEffect effect in mesh.Effects)
                {
                    effect.World = world;
                    effect.View = view;
                    effect.Projection = projection;
                }

                mesh.Draw();
            }
        }
    }
}
